import re
from typing import Optional

from . import models
from .config import config
from .models import ClientContext, ResponseContext


def get_context(client_name: str, /) -> Optional[ClientContext]:
    context: ClientContext
    for context in config.clients:
        if context.client_name.upper() == client_name.upper():
            return context

    return None


def fingerprint(data: dict, /) -> Optional[models.ResponseFingerprint]:
    response_context: Optional[ResponseContext] = get_response_context(data)

    if response_context is None:
        return None

    return models.ResponseFingerprint(
        request=response_context.request.type or None,
        function=response_context.function or None,
        browse_id=response_context.browse_id or None,
        context=response_context.context or None,
        client=response_context.client.name or None,
    )


def get_response_context(data: dict, /) -> Optional[ResponseContext]:
    response_context: Optional[dict] = data.get("responseContext")

    if response_context is None:
        return None

    services: dict = {}

    tracker: dict
    for tracker in response_context.get("serviceTrackingParams", ()):
        param: dict
        for param in tracker.get("params", ()):
            services.setdefault(tracker["service"], {})[param["key"]] = param["value"]

    request_type: Optional[str] = None
    request_id: Optional[str] = None

    key: str
    value: str
    for key, value in services.get("CSI", {}).items():
        match: Optional[re.Match] = re.match(r"Get(.+)_rid", key)

        if match is not None:
            request_type = match.group(1)
            request_id = value

    return ResponseContext(
        function=services.get("CSI", {}).get("yt_fn"),
        browse_id=services.get("GFEEDBACK", {}).get("browse_id"),
        context=services.get("GFEEDBACK", {}).get("context"),
        visitor_data=response_context.get("visitorData"),
        request=ResponseContext.Request(
            type=request_type,
            id=request_id,
        ),
        client=ResponseContext.Client(
            name=services.get("CSI", {}).get("c"),
            version=services.get("CSI", {}).get("cver"),
        ),
        flags=ResponseContext.Flags(
            logged_in=(value := services.get("GFEEDBACK", {}).get("logged_in"))
            and bool(int(value)),
        ),
    )


def error(error: dict, /) -> models.Error:
    return models.Error(
        code=error["code"], message=error["message"], reason=error["status"]
    )


def contextualise(client_context: ClientContext, data: dict) -> dict:
    data.setdefault("context", {}).setdefault("client", {}).update(
        client_context.context()
    )

    return data



import dataclasses
from typing import List, Optional

import httpx
import mediate
from httpx._types import ProxiesTypes

from . import api, utils
from .adaptor import InnerTubeAdaptor
from .config import config
from .enums import Endpoint
from .models import ClientContext, Locale
from .protocols import Adaptor


@dataclasses.dataclass
class Client:
    adaptor: Adaptor

    middleware: mediate.Middleware = dataclasses.field(
        default_factory=mediate.Middleware, repr=False, init=False
    )

    def __call__(
        self, endpoint: str, params: Optional[dict] = None, body: Optional[dict] = None
    ) -> dict:
        @self.middleware.bind
        def process(data: dict, /) -> dict:
            return data

        response: dict = process(
            self.adaptor.dispatch(endpoint, params=params, body=body)
        )

        response.pop("responseContext")

        return response


@dataclasses.dataclass(init=False)
class InnerTube(Client):
    def __init__(
        self,
        client_name: str,
        client_version: Optional[str] = None,
        *,
        api_key: Optional[str] = None,
        user_agent: Optional[str] = None,
        referer: Optional[str] = None,
        locale: Optional[Locale] = None,
        auto: bool = True,
        proxies: ProxiesTypes = None,
    ) -> None:
        if client_name is None:
            raise ValueError("Precondition failed: Missing client name")

        kwargs: dict = utils.filter(
            dict(
                client_name=client_name,
                client_version=client_version,
                api_key=api_key,
                user_agent=user_agent,
                referer=referer,
                locale=locale,
            )
        )

        context: ClientContext

        auto_context: Optional[ClientContext]
        if auto and (auto_context := api.get_context(client_name)):
            context = dataclasses.replace(auto_context, **kwargs)
        else:
            if client_version is None:
                raise ValueError("Precondition failed: Missing client version")

            context = ClientContext(**kwargs)

        super().__init__(
            adaptor=InnerTubeAdaptor(
                context=context,
                session=httpx.Client(base_url=config.base_url, proxies=proxies),
            )
        )

    def config(self) -> dict:
        return self(Endpoint.CONFIG)

    def guide(self) -> dict:
        return self(Endpoint.GUIDE)

    def player(self, video_id: str) -> dict:
        return self(
            Endpoint.PLAYER,
            body=dict(
                videoId=video_id,
            ),
        )

    def browse(
        self,
        browse_id: Optional[str] = None,
        *,
        params: Optional[str] = None,
        continuation: Optional[str] = None,
    ) -> dict:
        return self(
            Endpoint.BROWSE,
            body=utils.filter(
                dict(
                    browseId=browse_id,
                    params=params,
                    continuation=continuation,
                )
            ),
        )

    def search(
        self,
        query: Optional[str] = None,
        *,
        params: Optional[str] = None,
        continuation: Optional[str] = None,
    ) -> dict:
        return self(
            Endpoint.SEARCH,
            body=utils.filter(
                dict(
                    query=query or "",
                    params=params,
                    continuation=continuation,
                )
            ),
        )

    def next(
        self,
        video_id: Optional[str] = None,
        playlist_id: Optional[str] = None,
        *,
        params: Optional[str] = None,
        index: Optional[int] = None,
        continuation: Optional[str] = None,
    ) -> dict:
        return self(
            Endpoint.NEXT,
            body=utils.filter(
                dict(
                    params=params,
                    playlistId=playlist_id,
                    videoId=video_id,
                    playlistIndex=index,
                    continuation=continuation,
                )
            ),
        )

    def get_transcript(
        self,
        params: str,
    ) -> dict:
        return self(
            Endpoint.GET_TRANSCRIPT,
            body=utils.filter(
                dict(
                    params=params,
                )
            ),
        )

    def music_get_search_suggestions(
        self,
        input: Optional[None] = None,
    ) -> dict:
        return self(
            Endpoint.MUSIC_GET_SEARCH_SUGGESTIONS,
            body=dict(
                input=input or "",
            ),
        )

    def music_get_queue(
        self,
        *,
        video_ids: Optional[List[str]] = None,
        playlist_id: Optional[str] = None,
    ) -> dict:
        return self(
            Endpoint.MUSIC_GET_QUEUE,
            body=utils.filter(
                dict(
                    playlistId=playlist_id,
                    videoIds=video_ids or (None,),
                )
            ),
        )



from typing import Optional, Protocol, runtime_checkable


@runtime_checkable
class Adaptor(Protocol):
    def dispatch(
        self,
        endpoint: str,
        *,
        params: Optional[dict] = None,
        body: Optional[dict] = None
    ) -> dict:
        raise NotImplementedError



from typing import Optional

from httpx import Client, Request, Response

from . import api
from .config import config
from .errors import RequestError, ResponseError
from .models import ClientContext


class InnerTubeAdaptor:
    context: ClientContext
    session: Client

    def __init__(
        self, context: ClientContext, session: Optional[Client] = None
    ) -> None:
        self.context = context
        self.session = session or Client(base_url=config.base_url)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(context={self.context!r})"

    def _build_request(
        self, endpoint: str, params: Optional[dict] = None, body: Optional[dict] = None
    ) -> Request:
        return self.session.build_request(
            "POST",
            endpoint,
            params=self.context.params().update(params or {}),
            json=api.contextualise(self.context, body or {}),
            headers=self.context.headers(),
        )

    def _request(
        self, endpoint: str, params: Optional[dict] = None, body: Optional[dict] = None
    ) -> Response:
        return self.session.send(
            self._build_request(endpoint, params=params, body=body)
        )

    def dispatch(
        self, endpoint: str, params: Optional[dict] = None, body: Optional[dict] = None
    ) -> dict:
        response: Response = self._request(endpoint, params=params, body=body)

        content_type: Optional[str] = response.headers.get("Content-Type")

        if content_type is not None:
            if not content_type.lower().startswith("application/json"):
                raise ResponseError(f"Expected JSON response, got {content_type!r}")

        response_data: dict = response.json()

        visitor_data: Optional[str] = response_data.get("responseContext", {}).get(
            "visitorData"
        )

        if visitor_data is not None:
            self.session.headers["X-Goog-Visitor-Id"] = visitor_data

        error: Optional[dict] = response_data.get("error")

        if error is not None:
            raise RequestError(api.error(error))

        return response_data



import dataclasses

from . import models


@dataclasses.dataclass
class RequestError(Exception):
    error: models.Error

    def __str__(self) -> str:
        return str(self.error)


class ResponseError(Exception):
    pass



from .models import ClientContext, Config

REFERER_YOUTUBE: str = "https://www.youtube.com/"
REFERER_YOUTUBE_MOBILE: str = "https://m.youtube.com/"
REFERER_YOUTUBE_MUSIC: str = "https://music.youtube.com/"
REFERER_YOUTUBE_KIDS: str = "https://www.youtubekids.com/"
REFERER_YOUTUBE_STUDIO: str = "https://studio.youtube.com/"
REFERER_YOUTUBE_ANALYTICS: str = "https://analytics.youtube.com/"
REFERER_GOOGLE_ASSISTANT: str = "https://assistant.google.com/"

USER_AGENT_WEB: str = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.157 Safari/537.36"
USER_AGENT_ANDROID: str = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/65.0.3325.181 Mobile Safari/537.36"
USER_AGENT_IOS: str = "Mozilla/5.0 (iPhone; CPU iPhone OS 15_4_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/98.2  Mobile/15E148 Safari/605.1.15"
USER_AGENT_TV_HTML5: str = (
    "Mozilla/5.0 (PlayStation 4 5.55) AppleWebKit/601.2 (KHTML, like Gecko)"
)
USER_AGENT_TV_APPLE: str = (
    "AppleCoreMedia/1.0.0.12B466 (Apple TV; U; CPU OS 8_1_3 like Mac OS X; en_us)"
)
USER_AGENT_TV_ANDROID: str = "Mozilla/5.0 (Linux; Android 5.1.1; AFTT Build/LVY48F; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/49.0.2623.10"
USER_AGENT_XBOX_ONE: str = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox One) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Safari/537.36 Edge/13.10553"
USER_AGENT_GOOGLE_ASSISTANT: str = "Mozilla/5.0 (Linux; Android 11; Pixel 2; DuplexWeb-Google/1.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.193 Mobile Safari/537.36"

config: Config = Config(
    base_url="https://youtubei.googleapis.com/youtubei/v1/",
    clients=[
        ClientContext(
            client_id=1,
            client_name="WEB",
            client_version="2.20230728.00.00",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE,
            api_key="AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        ),
        ClientContext(
            client_id=2,
            client_name="MWEB",
            client_version="2.20211214.00.00",
            user_agent=USER_AGENT_ANDROID,
            referer=REFERER_YOUTUBE_MOBILE,
            api_key="AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        ),
        ClientContext(
            client_id=3,
            client_name="ANDROID",
            client_version="17.13.3",
            user_agent=USER_AGENT_ANDROID,
            api_key="AIzaSyA8eiZmM1FaDVjRy-df2KTyQ_vz_yYM39w",
        ),
        ClientContext(
            client_id=5,
            client_name="IOS",
            client_version="17.14.2",
            user_agent=USER_AGENT_IOS,
            api_key="AIzaSyB-63vPrdThhKuerbB2N_l7Kwwcxj6yUAc",
        ),
        ClientContext(
            client_id=7,
            client_name="TVHTML5",
            client_version="7.20210224.00.00",
            user_agent=USER_AGENT_TV_HTML5,
            api_key="AIzaSyDCU8hByM-4DrUqRUYnGn-3llEO78bcxq8",
        ),
        ClientContext(
            client_id=8,
            client_name="TVLITE",
            client_version="2",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=10,
            client_name="TVANDROID",
            client_version="1.0",
            user_agent=USER_AGENT_TV_ANDROID,
        ),
        ClientContext(
            client_id=13,
            client_name="XBOXONEGUIDE",
            client_version="1.0",
            user_agent=USER_AGENT_XBOX_ONE,
        ),
        ClientContext(
            client_id=14,
            client_name="ANDROID_CREATOR",
            client_version="21.06.103",
            user_agent=USER_AGENT_ANDROID,
            api_key="AIzaSyD_qjV8zaaUMehtLkrKFgVeSX_Iqbtyws8",
        ),
        ClientContext(
            client_id=15,
            client_name="IOS_CREATOR",
            client_version="20.47.100",
            user_agent=USER_AGENT_IOS,
            api_key="AIzaSyAPyF5GfQI-kOa6nZwO8EsNrGdEx9bioNs",
        ),
        ClientContext(
            client_id=16,
            client_name="TVAPPLE",
            client_version="1.0",
            user_agent=USER_AGENT_TV_APPLE,
        ),
        ClientContext(
            client_id=18,
            client_name="ANDROID_KIDS",
            client_version="7.12.3",
            user_agent=USER_AGENT_ANDROID,
            api_key="AIzaSyAxxQKWYcEX8jHlflLt2Qcbb-rlolzBhhk",
        ),
        ClientContext(
            client_id=19,
            client_name="IOS_KIDS",
            client_version="5.42.2",
            user_agent=USER_AGENT_IOS,
            api_key="AIzaSyA6_JWXwHaVBQnoutCv1-GvV97-rJ949Bc",
        ),
        ClientContext(
            client_id=21,
            client_name="ANDROID_MUSIC",
            client_version="5.01",
            user_agent=USER_AGENT_ANDROID,
            api_key="AIzaSyAOghZGza2MQSZkY_zfZ370N-PUdXEo8AI",
        ),
        ClientContext(
            client_id=23,
            client_name="ANDROID_TV",
            client_version="2.16.032",
            user_agent=USER_AGENT_TV_ANDROID,
        ),
        ClientContext(
            client_id=26,
            client_name="IOS_MUSIC",
            client_version="4.16.1",
            user_agent=USER_AGENT_IOS,
            api_key="AIzaSyBAETezhkwP0ZWA02RsqT1zu78Fpt0bC_s",
        ),
        ClientContext(
            client_id=27,
            client_name="MWEB_TIER_2",
            client_version="9.20220325",
            user_agent=USER_AGENT_ANDROID,
            referer=REFERER_YOUTUBE_MOBILE,
        ),
        ClientContext(
            client_id=28,
            client_name="ANDROID_VR",
            client_version="1.28.63",
            user_agent=USER_AGENT_ANDROID,
        ),
        ClientContext(
            client_id=29,
            client_name="ANDROID_UNPLUGGED",
            client_version="6.13",
            user_agent=USER_AGENT_ANDROID,
        ),
        ClientContext(
            client_id=30,
            client_name="ANDROID_TESTSUITE",
            client_version="1.9",
            user_agent=USER_AGENT_ANDROID,
        ),
        ClientContext(
            client_id=31,
            client_name="WEB_MUSIC_ANALYTICS",
            client_version="0.2",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE_ANALYTICS,
        ),
        ClientContext(
            client_id=33,
            client_name="IOS_UNPLUGGED",
            client_version="6.13",
            user_agent=USER_AGENT_IOS,
        ),
        ClientContext(
            client_id=38,
            client_name="ANDROID_LITE",
            client_version="3.26.1",
            user_agent=USER_AGENT_ANDROID,
        ),
        ClientContext(
            client_id=39,
            client_name="IOS_EMBEDDED_PLAYER",
            client_version="2.3",
            user_agent=USER_AGENT_IOS,
        ),
        ClientContext(
            client_id=41,
            client_name="WEB_UNPLUGGED",
            client_version="1.20220403",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE,
        ),
        ClientContext(
            client_id=42,
            client_name="WEB_EXPERIMENTS",
            client_version="1",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE,
        ),
        ClientContext(
            client_id=43,
            client_name="TVHTML5_CAST",
            client_version="1.1",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=55,
            client_name="ANDROID_EMBEDDED_PLAYER",
            client_version="17.13.3",
            user_agent=USER_AGENT_ANDROID,
            api_key="AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        ),
        ClientContext(
            client_id=56,
            client_name="WEB_EMBEDDED_PLAYER",
            client_version="1.20220413.01.00",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE,
            api_key="AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        ),
        ClientContext(
            client_id=57,
            client_name="TVHTML5_AUDIO",
            client_version="2.0",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=58,
            client_name="TV_UNPLUGGED_CAST",
            client_version="0.1",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=59,
            client_name="TVHTML5_KIDS",
            client_version="3.20220325",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=60,
            client_name="WEB_HEROES",
            client_version="0.1",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE,
        ),
        ClientContext(
            client_id=61,
            client_name="WEB_MUSIC",
            client_version="1.0",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE_MUSIC,
        ),
        ClientContext(
            client_id=62,
            client_name="WEB_CREATOR",
            client_version="1.20210223.01.00",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE_STUDIO,
            api_key="AIzaSyBUPetSUmoZL-OhlxA7wSac5XinrygCqMo",
        ),
        ClientContext(
            client_id=63,
            client_name="TV_UNPLUGGED_ANDROID",
            client_version="1.22.062.06.90",
            user_agent=USER_AGENT_TV_ANDROID,
        ),
        ClientContext(
            client_id=64,
            client_name="IOS_LIVE_CREATION_EXTENSION",
            client_version="17.13.3",
            user_agent=USER_AGENT_IOS,
        ),
        ClientContext(
            client_id=65,
            client_name="TVHTML5_UNPLUGGED",
            client_version="6.13",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=66,
            client_name="IOS_MESSAGES_EXTENSION",
            client_version="16.20",
            user_agent=USER_AGENT_IOS,
            api_key="AIzaSyDCU8hByM-4DrUqRUYnGn-3llEO78bcxq8",
        ),
        ClientContext(
            client_id=67,
            client_name="WEB_REMIX",
            client_version="1.20230724.00.00",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE_MUSIC,
            api_key="AIzaSyC9XL3ZjWddXya6X74dJoCTL-WEYFDNX30",
        ),
        ClientContext(
            client_id=68,
            client_name="IOS_UPTIME",
            client_version="1.0",
            user_agent=USER_AGENT_IOS,
        ),
        ClientContext(
            client_id=69,
            client_name="WEB_UNPLUGGED_ONBOARDING",
            client_version="0.1",
            user_agent=USER_AGENT_WEB,
        ),
        ClientContext(
            client_id=70,
            client_name="WEB_UNPLUGGED_OPS",
            client_version="0.1",
            user_agent=USER_AGENT_WEB,
        ),
        ClientContext(
            client_id=71,
            client_name="WEB_UNPLUGGED_PUBLIC",
            client_version="0.1",
            user_agent=USER_AGENT_WEB,
        ),
        ClientContext(
            client_id=72,
            client_name="TVHTML5_VR",
            client_version="0.1",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=74,
            client_name="ANDROID_TV_KIDS",
            client_version="1.16.80",
            user_agent=USER_AGENT_TV_ANDROID,
        ),
        ClientContext(
            client_id=75,
            client_name="TVHTML5_SIMPLY",
            client_version="1.0",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=76,
            client_name="WEB_KIDS",
            client_version="2.20220414.00.00",
            referer=REFERER_YOUTUBE_KIDS,
            user_agent=USER_AGENT_WEB,
            api_key="AIzaSyBbZV_fZ3an51sF-mvs5w37OqqbsTOzwtU",
        ),
        ClientContext(
            client_id=77,
            client_name="MUSIC_INTEGRATIONS",
            client_version="0.1",
        ),
        ClientContext(
            client_id=80,
            client_name="TVHTML5_YONGLE",
            client_version="0.1",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=84,
            client_name="GOOGLE_ASSISTANT",
            client_version="0.1",
            user_agent=USER_AGENT_GOOGLE_ASSISTANT,
        ),
        ClientContext(
            client_id=85,
            client_name="TVHTML5_SIMPLY_EMBEDDED_PLAYER",
            client_version="2.0",
            user_agent=USER_AGENT_TV_HTML5,
        ),
        ClientContext(
            client_id=87,
            client_name="WEB_INTERNAL_ANALYTICS",
            client_version="0.1",
            user_agent=USER_AGENT_WEB,
            referer=REFERER_YOUTUBE_ANALYTICS,
        ),
        ClientContext(
            client_id=88,
            client_name="WEB_PARENT_TOOLS",
            client_version="1.20220403",
            user_agent=USER_AGENT_WEB,
        ),
        ClientContext(
            client_id=89,
            client_name="GOOGLE_MEDIA_ACTIONS",
            client_version="0.1",
            user_agent=USER_AGENT_GOOGLE_ASSISTANT,
        ),
        ClientContext(
            client_id=90,
            client_name="WEB_PHONE_VERIFICATION",
            client_version="1.0.0",
            user_agent=USER_AGENT_WEB,
        ),
        ClientContext(
            client_id=92,
            client_name="IOS_PRODUCER",
            client_version="0.1",
            user_agent=USER_AGENT_IOS,
        ),
        ClientContext(
            client_id=93,
            client_name="TVHTML5_FOR_KIDS",
            client_version="7.20220325",
            user_agent=USER_AGENT_TV_HTML5,
        ),
    ],
)



from .adaptor import InnerTubeAdaptor
from .api import contextualise, error, fingerprint, get_context, get_response_context
from .clients import Client, InnerTube
from .config import config
from .enums import Endpoint, Request
from .models import (
    ClientContext,
    Config,
    Error,
    Locale,
    ResponseContext,
    ResponseFingerprint,
)
from .protocols import Adaptor



import enum


class StrEnum(str, enum.Enum):
    pass


class Endpoint(StrEnum):
    CONFIG: str = "config"
    GUIDE: str = "guide"
    PLAYER: str = "player"
    BROWSE: str = "browse"
    SEARCH: str = "search"
    NEXT: str = "next"
    GET_TRANSCRIPT: str = "get_transcript"
    MUSIC_GET_SEARCH_SUGGESTIONS: str = "music/get_search_suggestions"
    MUSIC_GET_QUEUE: str = "music/get_queue"


class Request(StrEnum):
    CONFIG: str = "Config"
    SEARCH: str = "Search"
    PLAYER: str = "Player"
    HOME: str = "Home"
    CHANNEL_PAGE: str = "ChannelPage"
    PLAYLIST: str = "Playlist"
    WATCH_NEXT: str = "WatchNext"
    BROWSE_HOME_PAGE: str = "BrowseHomePage"
    BROWSE_ARTIST_DETAIL_PAGE: str = "BrowseArtistDetailPage"
    BROWSE_ALBUM_DETAIL_PAGE: str = "BrowseAlbumDetailPage"
    BROWSE_PLAYLIST_DETAIL_PAGE: str = "BrowsePlaylistDetailPage"
    BROWSE_EXPLORE_PAGE: str = "BrowseExplorePage"
    BROWSE_NEW_RELEASES_PAGE: str = "BrowseNewReleasesPage"
    BROWSE_CHARTS_PAGE: str = "BrowseChartsPage"
    BROWSE_MOODS_AND_GENRES_PAGE: str = "BrowseMoodsAndGenresPage"
    BROWSE_MOODS_AND_GENRES_CATEGORY_PAGE: str = "BrowseMoodsAndGenresCategoryPage"
    MUSIC_GUIDE: str = "MusicGuide"
    MUSIC_QUEUE: str = "MusicQueue"
    MUSIC_WATCH_NEXT: str = "MusicWatchNext"
    MUSIC_SEARCH_SUGGESTIONS: str = "MusicSearchSuggestions"
    MOBILE_MAIN_APP_GUIDE: str = "MobileMainAppGuide"
    WEB_MAIN_APP_GUIDE: str = "WebMainAppGuide"


class ErrorStatus(StrEnum):
    FAILED_PRECONDITION: str = "FAILED_PRECONDITION"


class RequestContext(StrEnum):
    CHANNEL_CREATOR: str = "channel_creator"


class BrowseId(StrEnum):
    # Youtube Music
    MUSIC_EXPLORE: str = "FEmusic_explore"
    MUSIC_NEW_RELEASES: str = "FEmusic_new_releases"
    MUSIC_CHARTS: str = "FEmusic_charts"
    MUSIC_HOME: str = "FEmusic_home"
    MUSIC_MOODS_AND_GENRES: str = "FEmusic_moods_and_genres"
    MUSIC_MOODS_AND_GENRES_CATEGORY: str = "FEmusic_moods_and_genres_category"
    # YouTube
    WHAT_TO_WATCH: str = "FEwhat_to_watch"
    SHORTS: str = "FEshorts"
    LIBRARY: str = "FElibrary"
    # YouTube Kids
    KIDS_HOME: str = "FEkids_home"



import dataclasses
import http
from typing import Dict, List, Optional

from . import utils


@dataclasses.dataclass
class Locale:
    hl: str
    gl: Optional[str] = None

    def accept_language(self) -> str:
        return ",".join(item for item in (self.hl, self.gl) if item is not None)


@dataclasses.dataclass
class Error:
    code: int
    message: str
    reason: str

    def __str__(self) -> str:
        return f"{self.code} {self.status.phrase}: {self.message}"

    @property
    def status(self) -> http.HTTPStatus:
        return http.HTTPStatus(self.code)


@dataclasses.dataclass
class ClientContext:
    client_name: str
    client_version: str
    client_id: Optional[int] = None
    api_key: Optional[str] = None
    user_agent: Optional[str] = None
    referer: Optional[str] = None
    locale: Optional[Locale] = None

    def params(self) -> Dict[str, str]:
        return utils.filter(
            {
                "key": self.api_key,
                "alt": "json",
            }
        )

    def context(self) -> Dict[str, str]:
        return dict(
            clientName=self.client_name,
            clientVersion=self.client_version,
        )

    def headers(self) -> Dict[str, str]:
        return utils.filter(
            {
                "X-Goog-Api-Format-Version": "1",
                "X-YouTube-Client-Name": str(self.client_id),
                "X-YouTube-Client-Version": self.client_version,
                "User-Agent": self.user_agent,
                "Referer": self.referer,
                "Accept-Language": self.locale and self.locale.accept_language(),
            }
        )


@dataclasses.dataclass
class Config:
    base_url: str
    clients: List[ClientContext]


@dataclasses.dataclass
class ResponseContext:
    @dataclasses.dataclass
    class Request:
        type: Optional[str] = None
        id: Optional[str] = None

    @dataclasses.dataclass
    class Client:
        name: Optional[str] = None
        version: Optional[str] = None

    @dataclasses.dataclass
    class Flags:
        logged_in: Optional[bool] = None

    function: Optional[str] = None
    browse_id: Optional[str] = None
    context: Optional[str] = None
    visitor_data: Optional[str] = None
    client: Client = dataclasses.field(default_factory=Client)
    request: Request = dataclasses.field(default_factory=Request)
    flags: Flags = dataclasses.field(default_factory=Flags)


@dataclasses.dataclass
class ResponseFingerprint:
    request: Optional[str] = None
    function: Optional[str] = None
    browse_id: Optional[str] = None
    context: Optional[str] = None
    client: Optional[str] = None
